.welcome, .score {
  font-size:2em;
}

.gameover {
  font-size:3em;
}

.goal {
  font-size:1.2em;
}

.tip {
  font-size: 1em;
}
 
.play {
  font-size: 1.5em;
  padding-right:12px;
  padding-left:12px;
  padding-top:6px;
  padding-bottom:6px;
}
